@@ -1,4 +1,7 @@
+# -*- coding: utf-8 -*-
+
 # Copyright 2005 Lars Wirzenius (liw@iki.fi)
+# Copyright © 2011-2014 Andreas Beckmann (anbe@debian.org)
 #
 # This program is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
@@ -26,11 +29,14 @@ Lars Wirzenius <liw@iki.fi>
 
 import logging
 import os
+import random
+import stat
 import tempfile
 import time
 import UserDict
 import apt_pkg
 
+import piupartslib
 from piupartslib.dependencyparser import DependencyParser
 
 apt_pkg.init_system()
@@ -38,7 +44,7 @@ apt_pkg.init_system()
 
 def rfc822_like_header_parse(input):
     headers = []
-    while 1:
+    while True:
         line = input.readline()
         if not line or line in ["\r\n", "\n"]:
             break
@@ -48,6 +54,7 @@ def rfc822_like_header_parse(input):
             headers.append(line)
     return headers
 
+
 class Package(UserDict.UserDict):
 
     def __init__(self, headers):
@@ -58,8 +65,10 @@ class Package(UserDict.UserDict):
             self[name.strip()] = value.strip()
         self._parsed_deps = {}
         self._parsed_alt_deps = {}
-        self._rrdep_count = None
-        self._block_count = None
+        self.rrdep_cnt = None
+        self.block_cnt = None
+        self.waiting_cnt = None
+        self.rdep_chain_len = None
 
     def _parse_dependencies(self, header_name):
         if header_name in self._parsed_deps:
@@ -100,10 +109,10 @@ class Package(UserDict.UserDict):
                 vlist += self._parse_alternative_dependencies(header)
         return vlist
 
-    def prefer_alt_depends(self, header_name,dep_idx,dep):
+    def prefer_alt_depends(self, header_name, dep_idx, dep):
         if header_name in self:
             if header_name not in self._parsed_deps:
-                  self._parse_dependencies(header_name)
+                self._parse_dependencies(header_name)
             if self._parsed_deps[header_name][dep_idx]:
                 self._parsed_deps[header_name][dep_idx] = dep
 
@@ -114,39 +123,26 @@ class Package(UserDict.UserDict):
                 vlist += self._parse_dependencies(header)
         return vlist
 
-    def is_testable(self):
-        """Are we testable at all? Required aren't."""
-        return self.get("Priority", "") != "required"
-
-    def rrdep_count(self):
-        """Get the recursive dependency count, if it has been calculated"""
-        if self._rrdep_count == None:
-            raise Exception('Reverse dependency count has not been calculated')
-        return(self._rrdep_count)
-
-    def set_rrdep_count(self, val):
-        self._rrdep_count = val
-
-    def block_count(self):
-        """Get the number of packages blocked by this package"""
-        if self._block_count == None:
-            raise Exception('Block count has not been calculated')
-        return(self._block_count)
-
-    def set_block_count(self, val):
-        self._block_count = val
-
     def dump(self, output_file):
         output_file.write("".join(self.headers))
 
 
 class PackagesFile(UserDict.UserDict):
 
-    def __init__(self, input):
+    def __init__(self):
         UserDict.UserDict.__init__(self)
-        self._read_file(input)
+        self._urllist = []
 
-    def _read_file(self, input):
+    def load_packages_urls(self, urls, restrict_packages=None):
+        for url in urls:
+            logging.debug("Opening %s.*" % url)
+            (url, stream) = piupartslib.open_packages_url(url)
+            logging.debug("Fetching %s" % url)
+            self._read_file(stream, restrict_packages=restrict_packages)
+            stream.close()
+            self._urllist.append(url)
+
+    def _read_file(self, input, restrict_packages=None):
         """Parse a Packages file and add its packages to us-the-dict"""
         while True:
             headers = rfc822_like_header_parse(input)
@@ -158,8 +154,15 @@ class PackagesFile(UserDict.UserDict):
                 if apt_pkg.version_compare(p["Version"], q["Version"]) <= 0:
                     # there is already a newer version
                     continue
+            if restrict_packages is not None:
+                if p["Package"] not in restrict_packages:
+                    # unwanted package
+                    continue
             self[p["Package"]] = p
 
+    def get_urls(self):
+        return self._urllist
+
 
 class LogDB:
 
@@ -181,6 +184,16 @@ class LogDB:
         except AttributeError:
             pass
 
+    def bulk_load_dir(self, dirname):
+        try:
+            cache = self.exists_cache
+        except AttributeError:
+            self.exists_cache = {}
+            cache = self.exists_cache
+        for basename in os.listdir(dirname):
+            if basename.endswith(".log"):
+                cache[os.path.join(dirname, basename)] = True
+
     def open_file(self, pathname, mode):
         return file(pathname, mode)
 
@@ -207,7 +220,7 @@ class LogDB:
         # Let's make it follow the umask.
         umask = os.umask(0)
         os.umask(umask)
-        os.chmod(temp_name, 0666 & ~umask)
+        os.chmod(temp_name, 0o666 & ~umask)
 
         full_name = os.path.join(subdir, self._log_name(package, version))
         try:
@@ -225,6 +238,10 @@ class LogDB:
             self.remove_file(full_name)
         self._evict(full_name)
 
+    def stat(self, subdir, package, version):
+        full_name = os.path.join(subdir, self._log_name(package, version))
+        return os.stat(full_name)
+
 
 class LogfileExists(Exception):
 
@@ -240,7 +257,7 @@ class PackagesDB:
         "successfully-tested",
         "failed-testing",
         "cannot-be-tested",
-        "essential-required",
+        "essential-required",  # obsolete
         "waiting-to-be-tested",
         "waiting-for-dependency-to-be-tested",
         "dependency-failed-testing",
@@ -250,7 +267,7 @@ class PackagesDB:
         "unknown",
         "unknown-preferred-alternative",  # obsolete
         "no-dependency-from-alternatives-exists",  # obsolete
-        #"does-not-exist",  # can only happen as query result for a dependency
+        # "does-not-exist",  # can only happen as query result for a dependency
     ]
 
     _good_states = [
@@ -259,6 +276,7 @@ class PackagesDB:
     ]
 
     _obsolete_states = [
+        "essential-required",
         "circular-dependency",
         "unknown-preferred-alternative",
         "no-dependency-from-alternatives-exists",
@@ -286,7 +304,10 @@ class PackagesDB:
         self._packages = None
         self._in_state = None
         self._package_state = {}
+        self._dependency_databases = []
         self._recycle_mode = False
+        self._candidates_for_testing = None
+        self._rdeps = None
         self.set_subdirs(ok="pass", fail="fail", evil="untestable",
                          reserved="reserved", morefail=["bugged", "affected"],
                          recycle="recycle")
@@ -340,10 +361,31 @@ class PackagesDB:
     def get_mtime(self):
         return max([os.path.getmtime(sdir) for sdir in self._all])
 
-    def read_packages_file(self, input):
-        self._packages_files.append(PackagesFile(input))
+    def load_packages_urls(self, urls):
+        pf = PackagesFile()
+        pf.load_packages_urls(urls)
+        self._packages_files.append(pf)
         self._packages = None
 
+    def load_alternate_versions_from_packages_urls(self, urls):
+        # take version numbers (or None) from alternate URLs
+        pf2 = PackagesFile()
+        pf2.load_packages_urls(urls)
+        for package in self.get_all_packages():
+            if package["Package"] in pf2:
+                package["Version"] = pf2[package["Package"]]["Version"]
+            else:
+                package["Version"] = "None"
+
+    def get_urls(self):
+        urls = []
+        for pf in self._packages_files:
+            urls.extend(pf.get_urls())
+        return urls
+
+    def set_dependency_databases(self, dependency_databases=[]):
+        self._dependency_databases = list(dependency_databases)
+
     def _find_all_packages(self):
         if self._packages is None:
             self._packages = {}
@@ -367,10 +409,9 @@ class PackagesDB:
             more = more[1:]
             if dep not in deps:
                 deps.append(dep)
-                if dep in self._packages:
-                    more += self._packages[dep].dependencies()
-                elif dep in self._virtual_packages:
-                    more += self._packages[self._virtual_packages[dep][0]].dependencies()
+                dep_pkg = self.get_package(dep, recurse=True, resolve_virtual=True)
+                if dep_pkg is not None:
+                    more += dep_pkg.dependencies()
         return deps
 
     def _get_dependency_cycle(self, package_name):
@@ -382,12 +423,7 @@ class PackagesDB:
             more = more[1:]
             if dep not in deps:
                 deps.append(dep)
-                if dep in self._packages:
-                    dep_pkg = self._packages[dep]
-                elif dep in self._virtual_packages:
-                    dep_pkg = self._packages[self._virtual_packages[dep][0]]
-                else:
-                    dep_pkg = None
+                dep_pkg = self.get_package(dep, recurse=True, resolve_virtual=True)
                 if dep_pkg is not None and package_name in self._get_recursive_dependencies(dep_pkg):
                     circular.append(dep)
                     more += dep_pkg.dependencies()
@@ -402,9 +438,6 @@ class PackagesDB:
             return "failed-testing"
         if self._logdb.log_exists(package, [self._evil]):
             return "cannot-be-tested"
-        if not package.is_testable():
-            return "essential-required"
-
         return "unknown"
 
     def _compute_package_state(self, package):
@@ -444,7 +477,7 @@ class PackagesDB:
                                 prefer_alt = alternative
                                 prefer_alt_score = 2
                             elif prefer_alt_score < 1 and \
-                                 altdep_state in ["waiting-to-be-tested", "waiting-for-dependency-to-be-tested"]:
+                                    altdep_state in ["waiting-to-be-tested", "waiting-for-dependency-to-be-tested"]:
                                 prefer_alt = alternative
                                 prefer_alt_score = 1
                             elif prefer_alt_score < 0 and altdep_state == "unknown":
@@ -455,16 +488,15 @@ class PackagesDB:
                     if prefer_alt_score >= 0:
                         package.prefer_alt_depends(header, d, prefer_alt)
 
-        deps = package.dependencies()
+        dep_states = [(dep, self.get_best_package_state(dep))
+                      for dep in package.dependencies()]
 
-        for dep in deps:
-            dep_state = self.get_package_state(dep)
+        for dep, dep_state in dep_states:
             if dep_state in self._propagate_error_state:
                 return self._propagate_error_state[dep_state]
 
         testable = True
-        for dep in deps:
-            dep_state = self.get_package_state(dep)
+        for dep, dep_state in dep_states:
             if dep_state not in self._good_states:
                 testable = False
                 break
@@ -475,8 +507,7 @@ class PackagesDB:
         circular_deps = self._get_dependency_cycle(package["Package"])
         if package["Package"] in circular_deps:
             testable = True
-            for dep in deps:
-                dep_state = self.get_package_state(dep)
+            for dep, dep_state in dep_states:
                 if dep in circular_deps:
                     # allow any non-error dep_state on the cycle for testing
                     # (error states are handled by the error propagation above)
@@ -488,8 +519,7 @@ class PackagesDB:
             if testable:
                 return "waiting-to-be-tested"
 
-        for dep in deps:
-            dep_state = self.get_package_state(dep)
+        for dep, dep_state in dep_states:
             if dep_state in self._propagate_waiting_state:
                 return self._propagate_waiting_state[dep_state]
 
@@ -501,6 +531,9 @@ class PackagesDB:
 
         self._stamp = time.time()
 
+        for subdir in self._all:
+            self._logdb.bulk_load_dir(subdir)
+
         self._find_all_packages()
 
         self._package_state = {}
@@ -518,6 +551,9 @@ class PackagesDB:
             else:
                 self._in_state[state].append(package_name)
 
+        for db in self._dependency_databases:
+            db._compute_package_states()
+
         while todo:
             package_names = todo
             todo = []
@@ -551,6 +587,9 @@ class PackagesDB:
     def get_error_states(self):
         return [x for x in self._propagate_error_state.keys() if x in self._states]
 
+    def get_waiting_states(self):
+        return [x for x in self._propagate_waiting_state.keys() if x in self._states]
+
     def get_pkg_names_in_state(self, state):
         self._compute_package_states()
         return set(self._in_state[state])
@@ -559,45 +598,79 @@ class PackagesDB:
         self._find_all_packages()
         return name in self._packages
 
-    def get_package(self, name):
-        return self._packages[name]
+    def get_package(self, name, recurse=False, resolve_virtual=False):
+        self._find_all_packages()
+        if name in self._packages:
+            return self._packages[name]
+        if recurse:
+            for db in self._dependency_databases:
+                if db.has_package(name):
+                    return db.get_package(name)
+        if resolve_virtual:
+            providers = self.get_providers(name, recurse=recurse)
+            if providers:
+                return self.get_package(providers[0], recurse=recurse, resolve_virtual=False)
+        return None
 
-    def get_providers(self, name):
+    def get_version(self, name):
+        self._find_all_packages()
+        if name in self._packages:
+            return self._packages[name]["Version"]
+        return None
+
+    def get_source(self, name):
+        self._find_all_packages()
+        if name in self._packages:
+            return self.get_control_header(name, "Source")
+        return None
+
+    def get_providers(self, name, recurse=True):
+        self._find_all_packages()
+        providers = []
         if name in self._virtual_packages:
-            return self._virtual_packages[name]
-        return []
+            providers.extend(self._virtual_packages[name])
+        if recurse:
+            for db in self._dependency_databases:
+                providers.extend(db.get_providers(name, recurse=False))
+        return providers
 
     def get_all_packages(self):
         self._find_all_packages()
-        return self._packages
+        return self._packages.values()
+
+    def get_all_package_names(self):
+        self._find_all_packages()
+        return self._packages.keys()
 
     def get_control_header(self, package_name, header):
+        self._find_all_packages()
         if header == "Source":
-          # binary packages build from the source package with the same name
-          # don't have a Source header, so let's try:
-          try:
-            _source = self._packages[package_name][header]
-            # for binNMU the Source header in Packages files holds the version
-            # too, so we need to chop it of:
-            if " " in _source:
-              source, version = _source.split(" ")
-            else:
-              source = _source
-          except:
-            source = self._packages[package_name]["Package"]
-          return source
+            # binary packages build from the source package with the same name
+            # don't have a Source header, so let's try:
+            try:
+                _source = self._packages[package_name][header]
+                # for binNMU the Source header in Packages files holds the version
+                # too, so we need to chop it of:
+                if " " in _source:
+                    source, version = _source.split(" ")
+                else:
+                    source = _source
+            except:
+                source = self._packages[package_name]["Package"]
+            return source
         elif header == "Uploaders":
-          # not all (source) packages have an Uploaders header
-          uploaders = ""
-          try:
-            uploaders = self._packages[package_name][header]
-          except:
-            pass
-          return uploaders
+            # not all (source) packages have an Uploaders header
+            uploaders = ""
+            try:
+                uploaders = self._packages[package_name][header]
+            except:
+                pass
+            return uploaders
         else:
-          return self._packages[package_name][header]
+            return self._packages[package_name][header]
 
-    def get_package_state(self, package_name, resolve_virtual=True):
+    def get_package_state(self, package_name, resolve_virtual=True, recurse=True):
+        self._compute_package_states()
         if package_name in self._package_state:
             return self._package_state[package_name]
         if package_name in self._virtual_packages:
@@ -606,22 +679,86 @@ class PackagesDB:
                 return self._package_state[provider]
             else:
                 return "virtual"
+        if recurse:
+            for db in self._dependency_databases:
+                state = db.get_package_state(package_name, resolve_virtual=resolve_virtual, recurse=False)
+                if state != "does-not-exist":
+                    return state
+        if package_name in ["ia32-libs-i386", "ia32-libs-gtk-i386", "libnss-mdns-i386"]:
+            # HACK! these are arch=i386 packages needed on amd64
+            return "essential-required"
         return "does-not-exist"
 
+    def get_best_package_state(self, package_name, resolve_virtual=True, recurse=True):
+        package_state = self.get_package_state(package_name, resolve_virtual=resolve_virtual, recurse=recurse)
+        if package_state in self._good_states:
+            return package_state
+        providers = []
+        if resolve_virtual:
+            providers = self.get_providers(package_name, recurse=recurse)
+        if not providers:
+            return package_state
+        states = [self.get_package_state(name, resolve_virtual=False, recurse=recurse) for name in [package_name] + providers]
+        for state in self._good_states + self._propagate_waiting_state.keys() + self._propagate_error_state.keys():
+            if state in states:
+                return state
+        return package_state
+
+    def _get_package_weight(self, p):
+        # compute the priority of a package that needs testing
+        # result will be used as a reverse sorting key, so higher is earlier
+        waiting_count = self.waiting_count(p["Package"])
+        rdep_chain_len = self.rdep_chain_len(p["Package"])
+
+        if not self._recycle_mode:
+            return (
+                min(rdep_chain_len, waiting_count),
+                    waiting_count,
+            )
+
+        try:
+            statobj = self._logdb.stat(self._recycle, p["Package"], p["Version"])
+            ctime = statobj[stat.ST_CTIME]  # last inode modification = time of linking into recycle/
+            mtime = statobj[stat.ST_MTIME]
+        except OSError:
+            ctime = 0
+            mtime = 0
+
+        return (
+            min(rdep_chain_len, waiting_count),
+                waiting_count,
+                not self._logdb.log_exists(p, [self._ok]),  # prefer problematic logs
+                -ctime / 3600,  # prefer older, at 1 hour granularity to allow randomization
+                -mtime / 3600,  # prefer older, at 1 hour granularity to allow randomization
+        )
+
     def _find_packages_ready_for_testing(self):
-        return self.get_pkg_names_in_state("waiting-to-be-tested")
+        if self._candidates_for_testing is None:
+            self._candidates_for_testing = [self.get_package(pn)
+                                            for pn in self.get_pkg_names_in_state("waiting-to-be-tested")]
+            self._candidates_for_testing = [p for p in self._candidates_for_testing
+                                            if not self._logdb.log_exists(p, [self._reserved]) or
+                                            self._logdb.log_exists(p, [self._recycle])]
+            if len(self._candidates_for_testing) > 1:
+                tuples = [(self._get_package_weight(p), random.random(), p)
+                          for p in self._candidates_for_testing]
+                self._candidates_for_testing = [x[-1]
+                                                for x in sorted(tuples, reverse=True)]
+        return self._candidates_for_testing[:]
+
+    def _remove_unavailable_candidate(self, p):
+        self._candidates_for_testing.remove(p)
 
     def reserve_package(self):
         all_but_recycle = [x for x in self._all if x != self._recycle]
-        pset = self._find_packages_ready_for_testing()
-        while (len(pset)):
-            p = self.get_package(pset.pop())
+        for p in self._find_packages_ready_for_testing():
             if self._recycle_mode and self._logdb.log_exists(p, [self._recycle]):
                 for vdir in all_but_recycle:
                     if self._logdb.log_exists(p, [vdir]):
                         self._logdb.remove(vdir, p["Package"], p["Version"])
                         logging.info("Recycled %s %s %s" % (vdir, p["Package"], p["Version"]))
             if self._logdb.log_exists(p, all_but_recycle):
+                self._remove_unavailable_candidate(p)
                 continue
             if self._logdb.log_exists(p, [self._recycle]):
                 self._logdb.remove(self._recycle, p["Package"], p["Version"])
@@ -669,51 +806,102 @@ class PackagesDB:
         else:
             raise LogfileExists(self._evil, package, version)
 
-    def calc_rrdep_counts(self):
-        """Calculate recursive reverse dependency counts for Packages"""
+    def _get_rdep_dict(self):
+        """Return dict of one-level reverse dependencies by package"""
 
-        self._find_all_packages()       # populate _packages
+        if self._rdeps is None:
+
+            self._rdeps = {}
+
+            for pkg_name in self.get_all_package_names():
+                for dep in self.get_package(pkg_name).dependencies():
+                    dep_pkg = self.get_package(dep, recurse=True, resolve_virtual=True)
+
+                    if dep_pkg is not None:
+                        dep = dep_pkg["Package"]
+
+                    if not dep in self._rdeps:
+                        self._rdeps[dep] = set()
+                    self._rdeps[dep].add(pkg_name)
+
+        return self._rdeps
+
+    def _calc_rrdep_pkg_counts(self, pkg):
+
+        pkg_name = pkg['Package']
         self._compute_package_states()  # populate _package_state
+
+        # calc full recursive reverse dependency package set
+        rrdep_set = set()
+        rdeps = self._get_rdep_dict()
+        next_level = set([pkg_name])
+        chain_len = 0
+
+        while next_level:
+            chain_len += 1
+            rrdep_set |= next_level
+            new_pkgs = next_level
+            next_level = set([y for x in new_pkgs if x in rdeps for y in rdeps[x]])
+            next_level -= rrdep_set
+
+        rrdep_set.remove(pkg_name)
+
+        # calculate and set the metrics
+        pkg.rrdep_cnt = len(rrdep_set)
+
         error_states = self.get_error_states()
+        if self._package_state[pkg_name] in error_states:
+            block_list = [x for x in rrdep_set
+                          if self._package_state[x] in error_states]
+            pkg.block_cnt = len(block_list)
+        else:
+            pkg.block_cnt = 0
 
-        # create a reverse dependency dictionary.
-        # entries consist of a one-level list of reverse dependency package names,
-        # by package name
-        rdeps = {}
-        for pkg_name in self._packages.keys():
-            # use the Packages dependencies() method for a conservative count
-            for dep in self._packages[pkg_name].dependencies():
-                if dep in rdeps:
-                    rdeps[dep].append( pkg_name )
-                else:
-                    rdeps[dep] = [pkg_name]
+        waiting_states = self.get_waiting_states()
+        if self._package_state[pkg_name] in waiting_states:
+            waiting_list = [x for x in rrdep_set
+                            if self._package_state[x] in waiting_states]
+            pkg.waiting_cnt = len(waiting_list)
+        else:
+            pkg.waiting_cnt = 0
 
-        def recurse_rdeps( pkg_name, rdeps, rrdep_dict ):
-            """ Recurse through the reverse dep arrays to determine the recursive
-                dependency count for a package. rrdep_dict.keys() contains the
-                accumulation of rdeps encountered"""
+        pkg.rdep_chain_len = chain_len
 
-            # are there any rdeps for this package?
-            if pkg_name in rdeps:
-                for rdep in rdeps[pkg_name]:
-                    # break circular dependency loops
-                    if not rdep in rrdep_dict:
-                        rrdep_dict[rdep] = 1
-                        rrdep_dict = recurse_rdeps( rdep, rdeps, rrdep_dict )
+    def block_count(self, name):
+        pkg = self.get_package(name)
+        if pkg is None:
+            return -1
+        if pkg.block_cnt is None:
+            self._calc_rrdep_pkg_counts(pkg)
 
-            return rrdep_dict
+        return pkg.block_cnt
 
-        # calculate all of the rrdeps and block counts
-        for pkg_name in self._packages.keys():
-            rrdep_list = recurse_rdeps( pkg_name, rdeps, {} ).keys()
-            self._packages[pkg_name].set_rrdep_count( len(rrdep_list) )
+    def rrdep_count(self, name):
+        pkg = self.get_package(name)
+        if pkg is None:
+            return -1
+        if pkg.rrdep_cnt is None:
+            self._calc_rrdep_pkg_counts(pkg)
 
-            if self._package_state[pkg_name] in error_states:
-                block_list = [x for x in rrdep_list
-                              if self._package_state[x] in error_states]
-            else:
-                block_list = []
-            self._packages[pkg_name].set_block_count( len(block_list) )
+        return pkg.rrdep_cnt
+
+    def waiting_count(self, name):
+        pkg = self.get_package(name)
+        if pkg is None:
+            return -1
+        if pkg.waiting_cnt is None:
+            self._calc_rrdep_pkg_counts(pkg)
+
+        return pkg.waiting_cnt
+
+    def rdep_chain_len(self, name):
+        pkg = self.get_package(name)
+        if pkg is None:
+            return -1
+        if pkg.rdep_chain_len is None:
+            self._calc_rrdep_pkg_counts(pkg)
+
+        return pkg.rdep_chain_len
 
 
 # vi:set et ts=4 sw=4 :
